#   -*- coding: utf-8 -*-
#   Copyright 2019 Karellen, Inc. and contributors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import signal
from multiprocessing.forkserver import ForkServer as _ForkServer, SIGNED_STRUCT

from gevent.os import make_nonblocking, nb_read, nb_write

__implements__ = ["ForkServer", "_forkserver", "ensure_running", "_serve_one",
                  "get_inherited_fds", "connect_to_new_process",
                  "set_forkserver_preload", "read_signed", "write_signed"]
__target__ = "multiprocessing.forkserver"


class ForkServer(_ForkServer):
    def connect_to_new_process(self, fds):
        parent_r, parent_w = super().connect_to_new_process(fds)
        make_nonblocking(parent_r)
        make_nonblocking(parent_w)
        return parent_r, parent_w

    def ensure_running(self):
        super().ensure_running()
        make_nonblocking(self._forkserver_alive_fd)


def read_signed(fd):
    data = b''
    length = SIGNED_STRUCT.size
    while len(data) < length:
        s = nb_read(fd, length - len(data))
        if not s:
            raise EOFError('unexpected EOF')
        data += s
    return SIGNED_STRUCT.unpack(data)[0]


def write_signed(fd, n):
    msg = SIGNED_STRUCT.pack(n)
    while msg:
        nbytes = nb_write(fd, msg)
        if nbytes == 0:
            raise RuntimeError('should not get here')
        msg = msg[nbytes:]


def _serve_one(child_r, fds, unused_fds, handlers):
    from multiprocessing import resource_tracker, spawn
    # close unnecessary stuff and reset signal handlers
    signal.set_wakeup_fd(-1)
    for sig, val in handlers.items():
        signal.signal(sig, val)
    for fd in unused_fds:
        os.close(fd)

    (_forkserver._forkserver_alive_fd,
     resource_tracker._resource_tracker._fd,
     *_forkserver._inherited_fds) = fds
    make_nonblocking(_forkserver._forkserver_alive_fd)
    make_nonblocking(resource_tracker._resource_tracker._fd)
    for ifd in _forkserver._inherited_fds:
        make_nonblocking(ifd)

    # Run process object received over pipe
    parent_sentinel = os.dup(child_r)
    make_nonblocking(child_r)
    make_nonblocking(parent_sentinel)
    code = spawn._main(child_r, parent_sentinel)

    return code


_forkserver = ForkServer()
ensure_running = _forkserver.ensure_running
get_inherited_fds = _forkserver.get_inherited_fds
connect_to_new_process = _forkserver.connect_to_new_process
set_forkserver_preload = _forkserver.set_forkserver_preload



#   -*- coding: utf-8 -*-
#   Copyright 2019 Karellen, Inc. and contributors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from multiprocessing.util import spawnv_passfds as _spawnv_passfd, register_after_fork

from gevent.os import _watch_child
from gevent.threading import local

__implements__ = ["spawnv_passfds", "ForkAwareLocal", "get_command_line_gevent_preamble"]
__target__ = "multiprocessing.util"


def get_command_line_gevent_preamble():
    from gevent.monkey import saved
    from gevent import config
    from gevent._config import ImportableSetting
    from geventmp.monkey import GEVENT_SAVED_MODULE_SETTINGS

    prog = 'from gevent import monkey; monkey.patch_all(**%r); ' + \
           'from gevent import config; [setattr(config, k, v) for k, v in %r.items()]; '

    args = (saved[GEVENT_SAVED_MODULE_SETTINGS],
            {k: getattr(config, k) for k in dir(config)
             if not isinstance(config.settings[k], ImportableSetting)})

    return prog, args


def spawnv_passfds(path, args, passfds):
    launch_args = args[2] if len(args) >= 3 else None
    if launch_args and launch_args.startswith("from multiprocessing.forkserver import main;"):
        prog, prog_args = get_command_line_gevent_preamble()
        prog += "%s"
        launch_args = prog % (prog_args + (launch_args,))
        args[2] = launch_args

    cpid = _spawnv_passfd(path, args, passfds)
    _watch_child(cpid)
    return cpid


class ForkAwareLocal(local):
    def __init__(self):
        register_after_fork(self, lambda obj: obj.__dict__.clear())

    def __reduce__(self):
        return type(self), ()



#   -*- coding: utf-8 -*-
#   Copyright 2019 Karellen, Inc. and contributors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import sys
from multiprocessing import context
from multiprocessing import reduction
from multiprocessing import util
from multiprocessing.synchronize import SemLock as _SemLock, \
    Semaphore as _Semaphore, \
    BoundedSemaphore as _BoundedSemaphore, \
    Lock as _Lock, \
    RLock as _RLock, \
    SEMAPHORE, SEM_VALUE_MAX
from threading import get_ident

from gevent.hub import _get_hub_noargs as get_hub
from gevent.os import nb_read, nb_write, _read, ignored_errors
from gevent.timeout import with_timeout, Timeout

__implements__ = ["SemLock", "Semaphore", "BoundedSemaphore", "Lock", "RLock"]
__target__ = "multiprocessing.synchronize"

_BUF_ONE = (1).to_bytes(8, sys.byteorder, signed=False)


class SemLockEventFd:
    def __init__(self, kind, value, maxvalue, *, ctx=None):
        self.kind = kind
        self.maxvalue = maxvalue
        self.fd = os.eventfd(value, os.EFD_CLOEXEC | os.EFD_NONBLOCK | os.EFD_SEMAPHORE)
        util.debug(f"created semlock with fd {self.fd}")
        self._reset()

    def _reset(self):
        self._fd_path = f"/proc/{os.getpid()}/fdinfo/{self.fd}"
        self._semlock = self
        self.count = 0
        self.tid = None

    def __del__(self):
        try:
            os.close(self.fd)
        except OSError:
            pass

    def acquire(self, block=True, timeout=None):
        if self.kind != SEMAPHORE and self._is_mine():
            self.count += 1
            return True

        if block:
            if timeout is not None:
                try:
                    with_timeout(timeout, nb_read, self.fd, 8)
                except Timeout:
                    return False
            else:
                nb_read(self.fd, 8)
        else:
            try:
                _read(self.fd, 8)
            except OSError as e:
                if e.errno not in ignored_errors:
                    raise
                return False

        self.count += 1
        self.tid = get_ident()
        return True

    def release(self):
        if self.kind != SEMAPHORE:
            if not self._is_mine():
                raise AssertionError("attempt to release recursive lock not owned by thread")

            if self.count > 1:
                self.count -= 1
                return
        elif self.maxvalue != SEM_VALUE_MAX:
            count = self._get_value()
            if count >= self.maxvalue:
                raise ValueError("semaphore or lock released too many times")

        nb_write(self.fd, _BUF_ONE)
        self.count -= 1

    def _get_value(self):
        with open(self._fd_path, "rb") as f:
            for line in f:
                if line.startswith(b"eventfd-count:"):
                    return int(line[14:].strip())

    def _count(self):
        return self.count

    def _is_mine(self):
        return self.count > 0 and self.tid == get_ident()

    def _is_zero(self):
        return self._get_value() == 0

    def __enter__(self):
        return self.acquire()

    def __exit__(self, *args):
        self.release()

    def __getstate__(self):
        context.assert_spawning(self)
        df = reduction.DupFd(self.fd)
        return df, self.kind, self.maxvalue

    def __setstate__(self, state):
        df, kind, maxvalue = state
        fd = df.detach()
        util.debug(f'recreated blocker with fd {fd}')
        self.kind = kind
        self.maxvalue = maxvalue
        self.fd = fd
        self._reset()


class SemLockSem(_SemLock):
    def _make_methods(self):
        self._acquire = self._semlock.acquire
        self._release = self._semlock.release

    def acquire(self, *args, **kwargs):
        if self._semlock.kind != SEMAPHORE:
            return self._acquire(*args, **kwargs)
        return get_hub().threadpool.apply(self._acquire, args, kwargs)

    def release(self, *args, **kwargs):
        self._release(*args, **kwargs)

    def __enter__(self):
        if self._semlock.kind != SEMAPHORE:
            return super().__enter__()
        return get_hub().threadpool.apply(super().__enter__)


try:
    os.eventfd
    SemLock = SemLockEventFd
except AttributeError:
    SemLock = SemLockSem

Semaphore = type("Semaphore", (SemLock,), dict(_Semaphore.__dict__))
BoundedSemaphore = type("BoundedSemaphore", (Semaphore,), dict(_BoundedSemaphore.__dict__))
Lock = type("Lock", (SemLock,), dict(_Lock.__dict__))
RLock = type("RLock", (SemLock,), dict(_RLock.__dict__))



#   -*- coding: utf-8 -*-
#   Copyright 2019 Karellen, Inc. and contributors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sys
from multiprocessing import spawn, util

__implements__ = ["get_command_line"]
__target__ = "multiprocessing.spawn"


def get_command_line(**kwds):
    '''
    Returns prefix of command line used for spawning a child process
    '''
    if getattr(sys, 'frozen', False):
        return ([sys.executable, '--multiprocessing-fork'] +
                ['%s=%r' % item for item in kwds.items()])
    else:
        prog, args = util.get_command_line_gevent_preamble()

        prog += 'from multiprocessing.spawn import spawn_main; ' + \
                'spawn_main(%s);'

        args += (', '.join('%s=%r' % item for item in kwds.items()),)

        prog %= args

        opts = util._args_from_interpreter_flags()
        return [spawn._python_exe] + opts + ['-c', prog, '--multiprocessing-fork']



#   -*- coding: utf-8 -*-
#   Copyright 2019 Karellen, Inc. and contributors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import sys
from importlib import import_module

GEVENT_SAVED_MODULE_SETTINGS = "_gevent_saved_patch_all_module_settings"


def _patch_module(name,
                  items=None,
                  _warnings=None,
                  _patch_kwargs=None,
                  _notify_will_subscribers=True,
                  _notify_did_subscribers=True,
                  _call_hooks=True,
                  _package_prefix='gevent.'):
    from gevent.monkey.api import patch_module

    gevent_module = import_module(_package_prefix + name)
    target_module_name = getattr(gevent_module, '__target__', name)
    target_module = import_module(target_module_name)

    patch_module(target_module, gevent_module, items=items,
                 _warnings=_warnings, _patch_kwargs=_patch_kwargs,
                 _notify_will_subscribers=_notify_will_subscribers,
                 _notify_did_subscribers=_notify_did_subscribers,
                 _call_hooks=_call_hooks)

    # On Python 2, the `futures` package will install
    # a bunch of modules with the same name as those from Python 3,
    # such as `_thread`; primarily these just do `from thread import *`,
    # meaning we have alternate references. If that's already been imported,
    # we need to attempt to patch that too.

    # Be sure to keep the original states matching also.

    alternate_names = getattr(gevent_module, '__alternate_targets__', ())
    from gevent.monkey._state import saved  # TODO: Add apis for these use cases.
    for alternate_name in alternate_names:
        alternate_module = sys.modules.get(alternate_name)
        if alternate_module is not None and alternate_module is not target_module:
            saved.pop(alternate_name, None)
            patch_module(alternate_module, gevent_module, items=items,
                         _warnings=_warnings,
                         _notify_will_subscribers=False,
                         _notify_did_subscribers=False,
                         _call_hooks=False)
            saved[alternate_name] = saved[target_module_name]

    return gevent_module, target_module


def _patch_mp(will_patch_all):
    from gevent.monkey._state import saved
    geventmp_arg = will_patch_all.will_patch_module("geventmp")
    if geventmp_arg is None or geventmp_arg:
        _patch_module("_mp.3._mp_spawn", _package_prefix='geventmp.')
        _patch_module("_mp.3._mp_util", _package_prefix='geventmp.')
        _patch_module("_mp.3._mp_connection", _package_prefix='geventmp.')
        _patch_module("_mp.3._mp_synchronize", _package_prefix='geventmp.')
        _patch_module("_mp.3._mp_popen_fork", _package_prefix='geventmp.')
        _patch_module("_mp.3._mp_popen_forkserver", _package_prefix='geventmp.')
        _patch_module("_mp.3._mp_popen_spawn_posix", _package_prefix='geventmp.')
        _patch_module("_mp.3._mp_forkserver", _package_prefix='geventmp.')
        _patch_module("_mp.3._mp_resource_tracker", _package_prefix='geventmp.')
    saved[GEVENT_SAVED_MODULE_SETTINGS]["geventmp"] = True



#   -*- coding: utf-8 -*-
#   Copyright 2019 Karellen, Inc. and contributors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from io import BytesIO
from multiprocessing.connection import (reduce_connection,
                                        _ConnectionBase as __ConnectionBase,
                                        Connection as _Connection)

import multiprocessing.reduction
from gevent.os import make_nonblocking, nb_read, nb_write
from gevent.selectors import GeventSelector

__implements__ = ["_ConnectionBase", "Connection", "_WaitSelector"]
__target__ = "multiprocessing.connection"


class _ConnectionBase(__ConnectionBase):
    def __init__(self, handle, readable=True, writable=True):
        super(_ConnectionBase, self).__init__(handle, readable, writable)
        make_nonblocking(handle)


class Connection(_ConnectionBase, _Connection):
    _write = nb_write
    _read = nb_read

    def _send(self, buf, write=_write):
        return super()._send(buf, write)

    def _recv(self, size, read=_read):
        return super()._recv(size, read)


def dump(obj, file, protocol=None):
    from pickletools import dis, optimize
    import sys
    import traceback
    from multiprocessing.util import get_logger

    out = BytesIO()
    multiprocessing.reduction.ForkingPickler(out, protocol).dump(obj)
    out_bytes = bytes(out.getbuffer())
    out_bytes = optimize(out_bytes)
    file.write(out_bytes)

    logger = get_logger()

    for h in logger.handlers:
        h.flush()

    dis(out_bytes, out=sys.stderr)
    logger.debug("".join(traceback.format_stack()))


# multiprocessing.reduction.dump = dump
multiprocessing.reduction.register(Connection, reduce_connection)
_WaitSelector = GeventSelector



#   -*- coding: utf-8 -*-
#   Copyright 2020 Karellen, Inc. and contributors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from multiprocessing.resource_tracker import ResourceTracker as _ResourceTracker

from gevent.os import make_nonblocking, nb_write

__implements__ = ["ResourceTracker", "_resource_tracker", "ensure_running",
                  "register", "unregister", "getfd"]
__target__ = "multiprocessing.resource_tracker"


class ResourceTracker(_ResourceTracker):
    def ensure_running(self):
        super().ensure_running()
        make_nonblocking(self._fd)

    def _send(self, cmd, name, rtype):
        self.ensure_running()
        msg = '{0}:{1}:{2}\n'.format(cmd, name, rtype).encode('ascii')
        if len(name) > 512:
            # posix guarantees that writes to a pipe of less than PIPE_BUF
            # bytes are atomic, and that PIPE_BUF >= 512
            raise ValueError('name too long')
        nbytes = nb_write(self._fd, msg)
        assert nbytes == len(msg), "nbytes {0:n} but len(msg) {1:n}".format(
            nbytes, len(msg))


_resource_tracker = ResourceTracker()
ensure_running = _resource_tracker.ensure_running
register = _resource_tracker.register
unregister = _resource_tracker.unregister
getfd = _resource_tracker.getfd



#!/usr/bin/env python
#   -*- coding: utf-8 -*-

from setuptools import setup
from setuptools.command.install import install as _install

class install(_install):
    def pre_install_script(self):
        pass

    def post_install_script(self):
        pass

    def run(self):
        self.pre_install_script()

        _install.run(self)

        self.post_install_script()

if __name__ == '__main__':
    setup(
        name = 'geventmp',
        version = '0.0.10',
        description = 'Multiprocessing Gevent Extension',
        long_description = "===============================================\n GeventMP_ - Gevent_ Multiprocessing Extension\n===============================================\n\n.. image:: https://img.shields.io/gitter/room/karellen/Lobby?logo=gitter\n   :target: https://app.gitter.im/#/room/#karellen_Lobby:gitter.im\n   :alt: Gitter\n.. image:: https://img.shields.io/github/actions/workflow/status/karellen/geventmp/build.yml?branch=master\n   :target: https://github.com/karellen/geventmp/actions/workflows/build.yml\n   :alt: Build Status\n.. image:: https://img.shields.io/coveralls/github/karellen/geventmp/master?logo=coveralls\n   :target: https://coveralls.io/r/karellen/geventmp?branch=master\n   :alt: Coverage Status\n\n|\n\n.. image:: https://img.shields.io/pypi/v/geventmp?logo=pypi\n   :target: https://pypi.org/project/geventmp/\n   :alt: GeventMP Version\n.. image:: https://img.shields.io/pypi/pyversions/geventmp?logo=pypi\n   :target: https://pypi.org/project/geventmp/\n   :alt: GeventMP Python Versions\n.. image:: https://img.shields.io/pypi/dd/geventmp?logo=pypi\n   :target: https://pypi.org/project/geventmp/\n   :alt: GeventMP Downloads Per Day\n.. image:: https://img.shields.io/pypi/dw/geventmp?logo=pypi\n   :target: https://pypi.org/project/geventmp/\n   :alt: GeventMP Downloads Per Week\n.. image:: https://img.shields.io/pypi/dm/geventmp?logo=pypi\n   :target: https://pypi.org/project/geventmp/\n   :alt: GeventMP Downloads Per Month\n\n|\n\n.. warning::\n    HIC SUNT DRACONES!!!\n\n    This code is experimental (beta). There is some testing, but a lots of things are in flux, and\n    some platforms don't work at all.\n\n    You MAY try to use this in production with the understanding that this is a beta-quality software.\n\n    That said, this code may crash your server, bankrupt your company, burn your house down and be mean\n    to your puppy.\n\n    You've been warned.\n\nProblem\n=======\n\nDue to internal implementation, `multiprocessing` (`MP`) is unsafe to use with Gevent_ even when `monkey-patched`__.\nNamely, the use of OS semaphore primitives and inter-process IO in `MP` will cause the main\nloop to stall/deadlock/block (specific issue depends on the version of CPython).\n\n__ monkey_\n\nSolution\n========\nGeventMP_ (`Gee-vent Em-Pee`) is a gevent_ multiprocessing extension plugin for the `monkey-patching`__ subsystem.\nAs with the rest of the monkey patch subsystem the process is fairly clear:\n\n__ monkey_\n\n1. Identify all places where blocking occurs and where it may stall the loop.\n2. If blocking occurs on a file descriptor (`FD`), try to convert the file descriptor from blocking to non-blocking\n   (sockets/pipes/fifos, sometimes even files where, rarely, appropriate) and replace blocking IO functions with their\n   gevent_ non-blocking equivalents.\n3. If blocking occurs in a Python/OS primitive that does not support non-blocking access and thus cannot be geventized,\n   wrap all blocking access to that primitive with native thread-pool-based wrappers and call it a day (while fully\n   understanding that primitive access latency will increase and raw performance may suffer as a result).\n4. If you are really brave and have lots of free time on your hands, completely replace a standard blocking Python\n   non-`FD`-based primitive with implementation based on an `FD`-based OS primitive (e.g. POSIX semaphore =>\n   Linux `eventfd-based semaphore for kernels > 2.6.30`__).\n5. Due to launching of separate processes in `MP`, figure out how, when, and whether to `monkey-patch`__ spawned/forked\n   children and grandchildren.\n\n__ eventfd_\n\n__ monkey_\n\nInstallation\n============\nThe package is hosted on PyPi_.\n\nFor stable version:\n\n.. code-block:: bash\n\n  pip install geventmp\n\nFor unstable version:\n\n.. code-block:: bash\n\n  pip install --pre geventmp\n\n\nOnce installed, `GeventMP`_ will activate by default in the below stanza.\n\n.. code-block:: python\n\n   from gevent.monkey import patch_all\n   patch_all()\n\nIf you would like `GeventMP`_ to not activate by default, either do not install it or explicitly disable it:\n\n.. code-block:: python\n\n   from gevent.monkey import patch_all\n   patch_all(geventmp=False)\n\nThat's it - there are no other flags, settings, properties or config values so far.\n\nSupported Platforms\n===================\n\n.. note::\n    All claims of support may not be real at all. You're welcome to experiment. See warnings on top.\n\n* Linux and Darwin.\n* CPython 3.9, 3.10, 3.11, 3.12, 3.13\n* PyPy 3.9, 3.10\n\nTODO\n====\n1. Monkey patch Windows to the extent possible.\n2. Lots of applications use `Billiard <https://github.com/celery/billiard>`_ for multiprocessing instead of stock Python\n   package. Consider monkey patching Billiard if detected.\n\nContact Us\n==========\n\nPost feedback and issues on the `Bug Tracker`_, `Gitter`_,\nand `Twitter (@karelleninc)`_.\n\n.. _Gevent: https://github.com/gevent/gevent/\n.. _geventmp: https://github.com/karellen/geventmp\n.. _bug tracker: https://github.com/karellen/geventmp/issues\n.. _gitter: https://gitter.im/karellen/Lobby\n.. _twitter (@karelleninc): https://twitter.com/karelleninc\n.. _monkey: https://en.wikipedia.org/wiki/Monkey_patch\n.. _eventfd: https://linux.die.net/man/2/eventfd\n.. _pypi: https://pypi.org/project/geventmp/\n",
        long_description_content_type = 'text/x-rst; charset=UTF-8',
        classifiers = [
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
            'Programming Language :: Python :: 3.12',
            'Programming Language :: Python :: 3.13',
            'Programming Language :: Python :: 3.14',
            'Programming Language :: Python :: Implementation :: CPython',
            'Programming Language :: Python :: Implementation :: PyPy',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: POSIX',
            'Topic :: Internet',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Intended Audience :: Developers',
            'Development Status :: 4 - Beta'
        ],
        keywords = 'gevent multiprocessing mp monkey',

        author = 'Karellen, Inc.',
        author_email = 'supervisor@karellen.co',
        maintainer = 'Arcadiy Ivanov',
        maintainer_email = 'arcadiy@ivanov.biz',

        license = 'Apache-2.0',

        url = 'https://github.com/karellen/geventmp',
        project_urls = {
            'Bug Tracker': 'https://github.com/karellen/geventmp/issues',
            'Documentation': 'https://github.com/karellen/geventmp/',
            'Source Code': 'https://github.com/karellen/geventmp/'
        },

        scripts = [],
        packages = [
            'geventmp',
            'geventmp._mp',
            'geventmp._mp.3'
        ],
        namespace_packages = [],
        py_modules = [],
        entry_points = {
            'gevent.plugins.monkey.will_patch_all': ['geventmp = geventmp.monkey:_patch_mp']
        },
        data_files = [],
        package_data = {
            'geventmp': ['LICENSE']
        },
        install_requires = ['gevent>=25.8.2'],
        dependency_links = [],
        zip_safe = True,
        cmdclass = {'install': install},
        python_requires = '>=3.9',
        obsoletes = [],
    )



#   -*- coding: utf-8 -*-
#   Copyright 2019 Karellen, Inc. and contributors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from multiprocessing.semaphore_tracker import SemaphoreTracker as _SemaphoreTracker

from gevent.os import make_nonblocking, nb_write

__implements__ = ["SemaphoreTracker", "_semaphore_tracker", "ensure_running",
                  "register", "unregister", "getfd"]
__target__ = "multiprocessing.semaphore_tracker"


class SemaphoreTracker(_SemaphoreTracker):
    def ensure_running(self):
        super().ensure_running()
        make_nonblocking(self._fd)

    def _send(self, cmd, name):
        self.ensure_running()
        msg = '{0}:{1}\n'.format(cmd, name).encode('ascii')
        if len(name) > 512:
            # posix guarantees that writes to a pipe of less than PIPE_BUF
            # bytes are atomic, and that PIPE_BUF >= 512
            raise ValueError('name too long')
        nbytes = nb_write(self._fd, msg)
        assert nbytes == len(msg)


_semaphore_tracker = SemaphoreTracker()
ensure_running = _semaphore_tracker.ensure_running
register = _semaphore_tracker.register
unregister = _semaphore_tracker.unregister
getfd = _semaphore_tracker.getfd



#   -*- coding: utf-8 -*-
#   Copyright 2022 Karellen, Inc. and contributors
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from gevent.os import make_nonblocking
from multiprocessing.popen_fork import Popen as _Popen

__implements__ = ["Popen"]
__target__ = "multiprocessing.popen_fork"


class Popen(_Popen):
    def _launch(self, process_obj):
        self.sentinel = None
        try:
            super()._launch(process_obj)
        finally:
            if self.sentinel is not None:
                make_nonblocking(self.sentinel)
